"""Main module for comand line usege of hombre_tools"""
from os import scandir
import logging
import sqlparse
from hombre_tools.utils.utils import argument_parser

def jde_comment():
    """ adds comment to JDE source queries"""
    from hombre_tools.tools.jde_sql_decorate import comment as _comment, OPTIONS
    print(ARGS.path)
    OPTIONS['strip_comments'] = ARGS.strip_comments
    header = ARGS.header
    _copyright = f'Copyright Edwards Lifesciences {ARGS.year}, All rights reserved.'
    author = ARGS.author
    path = ARGS.path

    files = [entry for entry in scandir(path) if entry.is_file()]

    for file in files:
        with open(file.path) as in_file:
            print(file.path)
            new_name = file.path.split('.')[0] + '.sql'
            script = sqlparse.format(in_file.read(), **OPTIONS)
            _header = [f'/*\n{_copyright}',
                       f'AUTHOR:{author}',
                       f'file_name: {file.name.split(".")[0] +".sql"}',
                       '*/\n']
            
            header = '\n'.join(_header)
            script = header + _comment(script)
            new_name = file.path.split('.')[0] + '.sql'
            with open(new_name, 'w') as out_file:
                out_file.write(script)

def profile():
    """wrapper for data profile module"""
    from hombre_tools.db.db import sqlalchemy_engine, db_read_sql, ora_url
    from hombre_tools.tools.data_profiling import data_profile
    url = ora_url(ARGS) # TODO via
    engine = sqlalchemy_engine(ARGS,url)
    data_frame = db_read_sql(engine, ARGS.sql)
    ARGS.data_frame = data_frame
    if not data_frame.empty:
        data_profile(data_frame, ARGS.name, category=ARGS.category)
    else:
        logging.warning('No data')
            

if __name__ == '__main__':
    try:
        PARSER = argument_parser()
        ARGS = PARSER.parse_args()
        ACTION = ARGS.action
        globals()[ACTION]()
    except KeyError as msg:
        print(f'Error action  {msg} is not defined')
    
    print('Done')



"""
For each line with JDE table or
column object adds comment with JDE object description
Stanislav Vohnik
2019-09-06
"""
import re
from os import path
from pandas import HDFStore, DataFrame

# RAEDING JDE CATALOG
with HDFStore('/'.join(path.realpath(__file__).split('\\')[:-1]) +'/catalog/catalog_new.h5') as store:
    print(store.info())
    TABLES = store.get('tables')
    COLUMNS = store.get('columns')

CATALOG = ((TABLES, 'Table'),
           (COLUMNS, 'Field'))

OPTIONS = {'keyword_case':'upper',
           'reindent':True,
           'indent_width':4}

P = re.compile(r'(\w*)')

STOP_WORD = {'', 'select', 'from', 'as', 'where', 'group', 'by',
             'left', 'join', 'is', 'not', 'null', 'inner',
             'outer', 'case', 'when', 'else', 'then', 'end',
             'union', 'nolock', 'in', 'order'}

#creating index to speed _find
TABLES.reset_index()
TABLES.set_index('Table', inplace=True)
COLUMNS.reset_index()
COLUMNS.set_index('Field', inplace=True)

def _find(catalog, word):
    """
    finds description by df index
    """
    descr = None
    if word.lower() in STOP_WORD:
        return None

    try:
       _desc = catalog.loc[word]
    except:
       return None
    else:
       if isinstance(_desc, DataFrame):
           descr = _desc.Description.values[0]
       else:
            descr = _desc.Description

    return descr

def comment(script):
    """
    For each line with JDE table or
    column adds comment with JDE object description
    """
    sqls = []
    for sql in script.split(';'):
        lines = []
        for line in sql.splitlines():
            if '*/' in line:
                continue
            _comment = []
            for word in P.findall(line):
                for _catalog, _ in CATALOG:
                    descr = _find(_catalog, word)
                    if descr:
                        _comment.append(f'{word}:{descr}')
            if _comment:
                line += f" /* {', '.join(_comment)} */"
            lines.append(line)
        sqls.append('\n'.join(lines[:]))
    return ';\n'.join(sqls)



"""
Module fetchs JDE table catalog to pandas data frame
Stanislav Vohnik
2019-08-20
"""
from unicodedata import normalize
from pandas import read_html, DataFrame, concat
from requests_html import HTMLSession
from hombre_tools.utils.utils import run_in_separate_threads

def get_table_data(find, head, key, index=0, url='http://www.jdetables.com'):
    """
    response to dataframe
        parameters:
            url
            find string
            head: header 0/1
            key: index of a table on a html page
            index: index column
    """
    response = SESSION.get(url)
    table = response.html.find(find, first=False)[key].html
    data_frame = read_html(table, skiprows=0, header=head, index_col=index)[key]
    try:
        data_frame['Fk'] = f"{system}_{url.split('=')[-1]}"
        del data_frame["SQL Create Statement"]
        del data_frame['Unnamed: 0']
    except:
        pass
    return data_frame

TABLES = DataFrame()
COLUMNS = DataFrame()
SESSION = HTMLSession()

RESPONSE = SESSION.get('http://www.jdetables.com')
FIND = 'body > div.noprint select > option'
OPTIONS = RESPONSE.html.find(FIND, first=False)
SYSTEMS = [option.attrs['value'] for option in OPTIONS[1:]]
TITLES = [normalize("NFKD", code.full_text) for code in OPTIONS[1:]]

# test patameters
# SYSTEMS = ['01']
# TITLES = ['01: Address Book']

for e, system in enumerate(SYSTEMS):
    # getting tables for a system
    _find = 'body > div.middle > div.content > table'
    url = f'http://www.jdetables.com/?schema=920&system={system}'
    df = get_table_data(_find, 1, 0, url=url)
    tables = df.Table[:]
    df['system_name'] = TITLES[e]
    df['system'] = system
    TABLES = concat([TABLES, df])

    # getting table's columns within a system
    _find = 'body > div.middle form table'
    args = {"find":_find, "head":0, "key":0, "index":1}
    urls = [f'http://www.jdetables.com/?schema=920&system={system}&table={table}' for table in tables]
    print(f"{e+1}/{len(SYSTEMS)}, system_{system}, #tables: {len(tables)}")
    COLUMNS = concat([COLUMNS, run_in_separate_threads(urls, get_table_data, **args)])

"""You may need to save it into an excel files"""
# TABLES  # pandas frame with catalog
# TABLES.to_excel('jde_tables.xlsx')
# COLUMNS.to_excel('jde_columns.xlsx')



"""Module for fetching metadata from database along with roe counts"""
from operator import itemgetter
from sqlalchemy import MetaData, create_engine, Table

METADATA = MetaData()
HOST = 'AWPJDEDB3'
DB = 'JDE_PROD_REP'
DRV = 'ODBC+Driver+13+for+SQL+Server'
MYPATH = 'c:/Users/stanislav_vohnik/Documents/JDE Queries Current'

TABLES = []
TABLES = {'F40344',}

ENGINE = create_engine(f'mssql+pyodbc://@{HOST}/{DB}?trusted_connection=yes&driver={DRV}')

for table in TABLES:
    _schema = table.split('.')
    if len(_schema) == 2:
        schema, name = _schema
    else:
        schema, name = 'PRODDTA', table
    print(f'{schema}.{name}')
    _ = Table(name, METADATA, autoload=True, autoload_with=ENGINE, schema=schema)

SQL = ' union all \n'.join([f"select '{key}' as name, count(*) from {key}"
                            for key in METADATA.tables.keys()])
CON = ENGINE.connect()
RESULT = sorted(CON.execute(SQL).fetchall(), key=itemgetter(1))
CON.close()

print('\n'.join([f'{e}, {key[0]}: {key[1]} x {len(METADATA.tables[key[0]].columns)}'
                 for e, key in enumerate(RESULT)]))

# print(enumerate(RESULT.keys())



"""
Example of connecting to edwards  instance
Stanislav Vohnik 2019-08-21
"""
from pandas import read_sql
from  pandas_profiling import ProfileReport
from sqlalchemy import create_engine

HOST = 'AWPJDEDB3'
DB = 'JDE_PROD_REP'
DRIVER = 'ODBC+Driver+13+for+SQL+Server'
URL = f'mssql+pyodbc://@{HOST}/{DB}?trusted_connection=yes&driver={DRIVER}'

EMGINE = create_engine(URL)

TABLES = ['PRODDTA.F0025',
        'PRODDTA.F0013',
        'PRODDTA.F40205',
        'PRODDTA.F0010',
        'PRODDTA.F1113',
        'PRODDTA.F55XLVL2',
        'PRODDTA.F5541JP1',
        'PRODDTA.F0006',
        'PRODDTA.F01151',
        'PRODDTA.F4101D',
        'PRODDTA.F0007',
        'PRODDTA.F554701']

for table in  TABLES:
    df = read_sql(f"select * from {table}", EMGINE)
    report = ProfileReport(df)
    
    for col in df.filter(regex='^ID').columns:
        df[col] = df[col].astype('categorical')
    report.to_file(f'{table}_profile.html')
   
    print(f'saved: {table}')

# import HPFS5
# comand = 'bcp JDE_PROD_REP.PRODCTL.F0101 out F0101.csv  -b format -c'
# '''https://docs.microsoft.com/en-us/sql/tools/bcp-utility?view=sql-server-2017'''



"""
Example of connecting to edvards oracle instance
Stanislav Vohnik 2019-08-21
"""
from os import environ
from sqlalchemy import create_engine, exc
from keyring import get_password as pswd
import logging
from pandas import DataFrame, read_sql

def ora_url(args):
    return f'oracle://{args.user}:{pswd(args.host, args.user)}@{args.host}/{args.sid}'

def mssql_url(args, drv='ODBC+Driver+13+for+SQL+Server'):
    return f'mssql+pyodbc://@{args.host}/{args.db}?trusted_connection=yes&driver={drv}'

def sqlalchemy_engine(args, url):
    """engine constructor"""
    environ['PATH'] = args.ora_path  # we have to point to oracle client directory
    url = f'oracle://{args.user}:{pswd(args.host, args.user)}@{args.host}/{args.sid}'
    logging.info(url)
    return create_engine(url) # SQL Alchemy Engine

def get_ora_tables(engine):
    """wrapper for pandas read_sql returns data frame 'select * from all_tables' """
    sql = 'select * from all_tables'
    return read_sql(engine, sql)

def db_read_sql(engine, sql):
    """wrapper for pandas read_sql returns data frame"""
    data_frame = DataFrame()
    print(engine)
    try:
        with engine.connect() as con:
            data_frame = read_sql(sql, con)
    except exc.DatabaseError as msg:
        logging.error(msg)
    return data_frame # Pandas DataFrame


"""setup.py"""
from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md')) as f:
    long_description = f.read()

setup(
    name='hombre_tools',  # Required

    version='0.1.25',  # Required
    description='tools for daily usage',  # Optional
    url='https://github.com/hombre66/hombre_tools',  # Optional

    packages=find_packages(exclude=['contrib', 'docs', 'tests']),  # Required
    include_package_data=True,
    python_requires='>=3.6',
    install_requires=['numpy',
                      'pandas',
                      'pyOpenSSL',
                      'pyperclip',
                      'python-dateutil',
                      'python-settings',
                      'pywin32',
                      'pywin32-ctypes',
                      'pywinpty',
                      'requests',
                      'seaborn',
                      'SQLAlchemy',
                      'sqlparse',
                      'tables',
                      'urllib3',
                      'zipp',
                      'lxml'],
    extras_require={  # Optional
        'dev': ['check-manifest'],
        'test': ['coverage'],
    },
    data_files=[('hombre_tools/tools/catalog', ['hombre_tools/tools/catalog/catalog_new.h5'])], 

)



"""
wraper for data profileing module
"""
import pandas_profiling as pp
from time import asctime
import re

def data_profile(data_frame, name, category='(_key$)|(_id$)|(_number$)'):
    dtype = {col:'category'
            for col in data_frame.columns
                    if isinstance(data_frame[col].values[0], str) or re.findall(category, col.lower())}

    print(asctime(), '\tProfileReport')
    data_frame=data_frame.astype(dtype)
    report = pp.ProfileReport(data_frame)

    report.to_file(f"{name}_report.html")
    print(asctime(), f'Done {name}')



'''
set of tools utilites
'''
import concurrent.futures
import logging
from argparse import ArgumentTypeError, ArgumentParser
from pandas import DataFrame, concat as pd_concat


def str2bool(answer):
    """function for argparse to validate boolean input"""
    if answer.lower() in ('yes', 'true', 't', 'y', '1'):
        result = True
    elif answer.lower() in ('no', 'false', 'f', 'n', '0'):
        result = False
    else:
        raise ArgumentTypeError("Boolean value expected.")
    return result

def argument_parser(description='tools cmd'):
    """Common set of arguments for CMD tools"""
    description = '''Hombre tools\n\r
       Examples: \n\r
                python -m hombre_tools --action profile --sql "select * from dim_country" --user xxxx --host host --sid sid --name nameit\n\r
                python -m hombre_tools --action jde_comment  --path "C:/Queries"'''
    
    parser = ArgumentParser(description=description)

    parser.add_argument("--login",
                        help='db login',
                        required=False)

    parser.add_argument("--host",
                        required=False,
                        help='uri of host')

    parser.add_argument("--path",
                        default=None,
                        required=False)

    parser.add_argument("--strip_comments",
                        const=False,
                        default=False,
                        type=str2bool,
                        nargs='?',
                        required=False,
                        help="If True existing comments are removed from the statements")

    parser.add_argument("--author", default='', required=False)

    parser.add_argument("--header", default='JDE Queries', required=False)

    parser.add_argument("--year", default='2019', required=False)

    parser.add_argument("--action", required=False, help='jde_comment, profile')


    parser.add_argument("--category",
                        required=False,
                        default='(_key$)|(_id$)|(_number$)')

    parser.add_argument("--name", required=False)

    parser.add_argument("--sql", required=False)

    parser.add_argument("--user", required=False)

    parser.add_argument("--ora_path",
                        required=False,
                        default='C:/oracle/instantclient;%PATH%')

    parser.add_argument("--sid", required=False)


    return parser

def max_lenght(data_frame, col):
    '''returns leghth of fdataframe object'''
    return max(len(data_frame[col].astype(str).min()), len(data_frame[col].astype(str).max())) + 1


def run_in_separate_threads(items, target, **args):
    """Wrapper to execute method in separate tread
    for each "item" and concat them into DataFrame."""
    _result = DataFrame()

    with concurrent.futures.ThreadPoolExecutor(max_workers=100) as executor:
        future_result = {executor.submit(target, *list(args.values()) + [item]):\
                         e for e, item in enumerate(items)}
        for future in concurrent.futures.as_completed(future_result):
            try:
                data = future.result()
            except (ValueError, AttributeError) as exc:
                logging.warning(' %s generated an exception: %s', target.__name__, exc)
                data = DataFrame()

            _result = pd_concat([_result, data])

    return _result



"""
stanislav Vohnik password utility
2019-08-10
"""
import keyring
from argparse import ArgumentParser
from pyperclip import paste, copy as clipcopy

PARSER = ArgumentParser(description='Password Utility version 0.1')

PARSER.add_argument("--user",
                    help='login name',
                    required=False,
                    default='Stanislav_Vohnik')

PARSER.add_argument("--service",
                    help='service name',
                    required=True)

PARSER.add_argument("--action",
                    help='login name',
                    required=True)

if __name__ == 'main':
    ARGS = PARSER.parse_args()
    if ARGS.action == 'set':
        pswd = paste
        keyring.set_password(ARGS.sevice, ARGS.user, pswd)
        print('password set from clip-board')

    if ARGS.action == 'get':
        clipcopy(keyring.get_password(ARGS.sevice, ARGS.user))
        print('get to clip board')
        